from ngsolve import *
from netgen.csg import *
import netgen.gui

distances= [13.8] #distance of the magnet for <20[mm] it is inside the coil
j=(1 * 700)/0.00005 #current density (I*N)/A

forces = distances[:]
w=[None, None]
distance_magnet = distances[len(distances)-1]/1000
box = OrthoBrick(Pnt(-0.090, -0.090, -0.090),Pnt(distance_magnet+0.105, 0.1246, 0.1084)).bc("outer")
coil1 = (OrthoBrick(Pnt(0, 0,0),Pnt(0.020, 0.0346, 0.0025)))
coil2 = (OrthoBrick(Pnt(0, 0, 0.0025),Pnt(0.020, 0.0025, 0.0159)))
coil3 = (OrthoBrick(Pnt(0, 0, 0.0159),Pnt(0.020, 0.0346, 0.0184)))
coil4 = (OrthoBrick(Pnt(0, 0.0321, 0.0025),Pnt(0.020, 0.0346, 0.0159)))

for x in range(0, len(distances)):
    distance_magnet = distances[x]/1000
    magnet = OrthoBrick(Pnt(distance_magnet, 0.0048, 0.0035),Pnt(distance_magnet+0.015, 0.0298, 0.0145)).bc('gamma')
    air = box - magnet - coil1 - coil2 - coil3 - coil4

    geo = CSGeometry()
    geo.Add (air.mat("air"))
    geo.Add (coil1.mat("coil1").maxh(0.005))
    geo.Add (coil2.mat("coil2").maxh(0.005))
    geo.Add (coil3.mat("coil3").maxh(0.005))
    geo.Add (coil4.mat("coil4").maxh(0.005))
    geo.Add (magnet.mat("magnet").maxh(0.003))
    mesh = Mesh(geo.GenerateMesh(maxh=0.030))
    mesh.GetMaterials(), mesh.GetBoundaries()
    fes = HCurl(mesh, order=3, dirichlet="outer", flags = { "nograds" : True })
    print ("ndof =", fes.ndof)
    u = fes.TrialFunction()
    v = fes.TestFunction()


    from math import pi
    mu0 = 4*pi*1e-7
    mur = CoefficientFunction( [1.0445730167132 if mat== "magnet" else #myr magnet
                                1.0000000000000 if mat== "air" else 0.999991 #myr Kupfer
                                for mat in mesh.GetMaterials()]) 

    a = BilinearForm(fes)
    a += SymbolicBFI(1/(mu0*mur) * curl(u) * curl(v) + 1e-6/(mu0*mur)*u*v)
    c = Preconditioner(a, "bddc")

    f = LinearForm(fes)
    mag = CoefficientFunction((-837999.999999998,0,0)) #Coercivity
    f += SymbolicLFI(mag*curl(v), definedon=mesh.Materials("magnet"))
    f += SymbolicLFI(CoefficientFunction((0,j,0)) * v, definedon=mesh.Materials("coil1"))
    f += SymbolicLFI(CoefficientFunction((0,0,-j)) * v, definedon=mesh.Materials("coil2"))
    f += SymbolicLFI(CoefficientFunction((0,-j,0)) * v, definedon=mesh.Materials("coil3"))
    f += SymbolicLFI(CoefficientFunction((0,0,j)) * v, definedon=mesh.Materials("coil4"))
 
    with TaskManager():
        a.Assemble()
        f.Assemble()

    gfu = GridFunction(fes)

    inv = CGSolver(a.mat, c.mat)
    gfu.vec.data = inv * f.vec
    Draw (curl(gfu), mesh, "B-field")
    print('B-field simulation done')

    ################
   
    print('starting force calculation...')
    B = curl(gfu)
    nu=1.0/(mu0*mur)
    # Maxwell stress tensor ...
    sigma = nu * (OuterProduct(B, B) - 0.5 * (B*B) * Id(3) )

    # compute force in H1-space:
    # compute f = div sigma  in weak form
    fes = VectorH1(mesh, order=3)
    u,v = fes.TnT()
    force = LinearForm(fes)
    force += SymbolicLFI(InnerProduct(sigma, grad(v)))
    force.Assemble()

    mass = BilinearForm(fes)
    mass += SymbolicBFI(u*v, BND)
    mass.Assemble()

    gfforce = GridFunction(fes)
    gfforce.vec.data = mass.mat.Inverse(inverse="sparsecholesky", freedofs=fes.GetDofs(mesh.Boundaries("gamma"))) * force.vec

    char_core = CoefficientFunction( [1 if bc=="gamma" else 0 for bc in mesh.GetBoundaries()] )

    force=Integrate(gfforce*char_core, mesh, BND)

    print ("facetforce-h1, core: ", force)
    forces[x]=force

print("Distance [mm]: Kraft F [N] x,y,z\n")
for x in range(0, len(distances)):
    print(distances[x], ":\n", forces[x], "\n")